"""
General utilities
"""
from datetime import date, timedelta, datetime
from functools import cache
from io import BytesIO
from logging import info
from typing import Dict, Optional, List
from zipfile import ZipFile

from pandas import concat
from pandera import check_types
from pandera.typing import DataFrame
from requests import get

from .abstractions import HasURL, HasZip, TimeUnit
from .concepts import SimpleFuel
from .config import TZ
from .schema import Snapshot, Summary

LABEL: Dict[SimpleFuel, str] = {
    SimpleFuel.FOSSIL_FUEL: "Fossil fuel",
    SimpleFuel.RENEWABLE: "Renewable",
    SimpleFuel.NUCLEAR: "Nuclear",
}


@cache
def get_last_modified(obj: HasURL) -> str:
    info(f"getting last modified for {obj}")
    with get(obj.url) as r:
        headers = r.headers
    return headers["Last-Modified"]


def get_zip(obj: HasZip, last_modified: Optional[str] = None) -> bytes:
    return obj.zip


def get_content(
    name_file_csv: str, zip: bytes, last_modified: Optional[str] = None
) -> bytes:
    with BytesIO(zip) as f:
        with ZipFile(f) as zf:
            content = zf.read(name_file_csv)
    return content


def get_name_field_gen_mw(df: DataFrame[Snapshot]) -> str:
    return set(df.columns).intersection({Snapshot.gen_mw, Snapshot.gen_mwh}).pop()


@check_types
def summarize(df: DataFrame[Snapshot]) -> DataFrame[Summary]:
    df[Summary.date] = df[Snapshot.timestamp].dt.date
    return (
        df.rename(columns={get_name_field_gen_mw(df): Summary.gen_mw})
        .groupby([Summary.date, Snapshot.fuel], as_index=False)
        .agg({Summary.gen_mw: "sum"})
    )


@check_types
def collect_summaries(
    time_units: List[TimeUnit], last_modified: Optional[str] = None
) -> DataFrame[Summary]:
    return concat([time_unit.summary for time_unit in time_units])


def get_end_of_last_month() -> date:
    now = datetime.now(TZ)
    return date(year=now.year, month=now.month, day=1) - timedelta(days=1)



"""
base classes in the nysmix data
"""

from datetime import date, timedelta
from functools import cached_property

from requests import get
from nysmix.config import DATE_START

from abc import ABC, abstractmethod, abstractproperty
from dataclasses import dataclass
from .config import FORMAT_DATE
from logging import info
from .utils import get_last_modified, get_end_of_last_month


@dataclass(frozen=True)  # type: ignore
class TimeUnitBase(ABC):
    year: int
    month: int
    day: int = 1

    @abstractproperty
    def last_modified(self) -> str:
        pass

    @property
    def dt(self) -> date:
        return date(year=self.year, month=self.month, day=self.day)

    @property
    def str_date(self) -> str:
        return self.dt.strftime(FORMAT_DATE)

    @property
    def is_before_start(self) -> bool:
        return self.dt < self.from_date(DATE_START).dt

    @property
    def is_future(self) -> bool:
        return self.dt > get_end_of_last_month()

    @staticmethod
    @abstractmethod
    def from_date(dt: date):  # -> TimeUnit:
        pass


class MonthBase(TimeUnitBase):
    @property
    def name_file_zip(self) -> str:
        return f"{self.str_date}rtfuelmix_csv.zip"

    @property
    def url(self) -> str:
        return f"http://mis.nyiso.com/public/csv/rtfuelmix/{self.name_file_zip}"

    @cached_property
    def last_modified(self) -> str:
        return get_last_modified(obj=self)

    @staticmethod
    def from_date(dt: date):  # -> TimeUnit:
        return MonthBase(year=dt.year, month=dt.month)

    @property
    def zip(self) -> bytes:
        info(f"attempt to download {self.url}")
        with get(self.url) as r:
            content = r.content
        return content


class DayBase(TimeUnitBase):
    @property
    def next(self):  # -> Day
        return DayBase.from_date(dt=self.dt + timedelta(days=1))

    @cached_property
    def yearmonth(self) -> MonthBase:
        return MonthBase(year=self.year, month=self.month)

    @property
    def name_file_csv(self) -> str:
        return f"{self.str_date}rtfuelmix.csv"

    @cached_property
    def last_modified(self) -> str:
        return self.yearmonth.last_modified

    @staticmethod
    def from_date(dt: date):  # -> Day:
        return DayBase(year=dt.year, month=dt.month, day=dt.day)



"""
general concepts
"""

from enum import Enum, auto
from typing import Dict, Set


class Fuel(Enum):
    HYDRO = auto()
    WIND = auto()
    OTHER_RENEWABLES = auto()
    NUCLEAR = auto()
    NATURAL_GAS = auto()
    DUAL_FUEL = auto()
    OTHER_FOSSIL_FUELS = auto()


class SimpleFuel(Enum):
    FOSSIL_FUEL = auto()
    RENEWABLE = auto()
    NUCLEAR = auto()


MEMBERS: Dict[SimpleFuel, Set[Fuel]] = {
    SimpleFuel.FOSSIL_FUEL: {Fuel.NATURAL_GAS, Fuel.DUAL_FUEL, Fuel.OTHER_FOSSIL_FUELS},
    SimpleFuel.RENEWABLE: {Fuel.HYDRO, Fuel.WIND, Fuel.OTHER_RENEWABLES},
    SimpleFuel.NUCLEAR: {Fuel.NUCLEAR},
}

NAME: Dict[Fuel | SimpleFuel, str] = {
    Fuel.HYDRO: "Hydro",
    Fuel.WIND: "Wind",
    Fuel.OTHER_RENEWABLES: "Other Renewables",
    Fuel.NUCLEAR: "Nuclear",
    Fuel.NATURAL_GAS: "Natural Gas",
    Fuel.DUAL_FUEL: "Dual Fuel",
    Fuel.OTHER_FOSSIL_FUELS: "Other Fossil Fuels",
    SimpleFuel.FOSSIL_FUEL: "Fossil fuel",
    SimpleFuel.NUCLEAR: "Nuclear",
    SimpleFuel.RENEWABLE: "Renewable",
}



from joblib import Memory
from pydantic import BaseSettings


class SettingsGCS(BaseSettings):
    name_bucket: str
    name_project: str
    name_folder_cache: str = "cache"
    verbosity: int = 0

    class Config:
        env_prefix = "gcs_"

    @property
    def memory(self) -> Memory:
        return Memory(
            f"{self.name_bucket}/{self.name_folder_cache}",
            backend="gcs",
            verbose=self.verbosity,
            backend_options={"project": self.name_project},
        )


"""
data implementations that optionally uses joblib caching
"""

from dataclasses import dataclass
from datetime import datetime
from functools import cached_property
from io import BytesIO
from logging import info
from typing import List, Optional

from pandas import concat, read_csv
from pandera import check_types
from pandera.typing import DataFrame

from .base import DayBase, MonthBase
from .config import TZ, YEAR_START, DATE_START
from .schema import Summary
from .utils import get_content, get_zip, summarize, collect_summaries
from .abstractions import TimeUnit
from abc import abstractproperty


class HasSubunits(TimeUnit):
    @abstractproperty
    def subunits(self) -> List[TimeUnit]:
        pass

    @property
    def last_modified(self) -> str:
        return self.subunits[-1].last_modified

    @property # type: ignore
    @check_types
    def summary(self) -> DataFrame[Summary]:
        return self.memory.cache(collect_summaries)(
            time_units=self.subunits, last_modified=self.last_modified
        )


@dataclass(frozen=True)
class Day(TimeUnit):
    day_base: DayBase = DayBase.from_date(DATE_START)

    @property
    def last_modified(self) -> str:
        return self.day_base.last_modified

    @cached_property
    def content(self) -> bytes:
        return self.memory.cache(get_content)(
            name_file_csv=self.day_base.name_file_csv,
            zip=self.month.zip,
            last_modified=self.last_modified,
        )

    @cached_property
    def df(self) -> DataFrame:
        with BytesIO(self.content) as f:
            df = read_csv(f, dtype=object)
        return df

    @property
    def summary(self) -> DataFrame[Summary]:
        info(f"summarizing {self}")
        return self.memory.cache(summarize)(df=self.df)

    @property
    def month(self):  # -> Month
        return Month(
            month_base=MonthBase.from_date(self.day_base.dt), memory=self.memory
        )


@dataclass(frozen=True)
class Month(HasSubunits):
    month_base: MonthBase = MonthBase.from_date(DATE_START)

    @property
    def subunits(self) -> List[TimeUnit]:
        days = []
        current = DayBase.from_date(dt=self.month_base.dt)
        while current.yearmonth == self.month_base:
            if not current.is_before_start and not current.is_future:
                days.append(current)
            current = current.next
        return [Day(day_base=day, memory=self.memory) for day in days]

    @property
    def last_modified(self) -> str:
        return self.month_base.last_modified

    @property
    def zip(self) -> bytes:
        return self.memory.cache(get_zip)(
            obj=self.month_base, last_modified=self.last_modified
        )


@dataclass(frozen=True)
class Year(HasSubunits):
    year: int = YEAR_START

    @property
    def subunits(self) -> List[TimeUnit]:
        return [
            Month(memory=self.memory, month_base=month)
            for month in [
                MonthBase(year=self.year, month=month) for month in range(1, 13)
            ]
            if not month.is_future and not month.is_before_start
        ]


@dataclass(frozen=True)
class All(HasSubunits):
    @property
    def subunits(self) -> List[TimeUnit]:
        now = datetime.now(TZ)
        return [
            Year(memory=self.memory, year=year)
            for year in range(YEAR_START, now.year + 1)
        ]



"""general configuration
"""

from enum import Enum
from typing import NamedTuple, Optional
from os import getenv
from datetime import date
from datetime import date, timedelta, timezone


YEAR_START = 2015
MONTH_START = 12
DAY_START = 9
DATE_START = date(year=YEAR_START, month=MONTH_START, day=DAY_START)


FORMAT_DATE = "%Y%m%d"
TZ = timezone(timedelta(hours=-5))

class ParamsVarEnv(NamedTuple):
    key: str

    @property
    def value(self) -> Optional[str]:
        return getenv(self.key)


class VarEnv(ParamsVarEnv, Enum):
    TYPE_STORE = ParamsVarEnv(key="NYSMIX_STORE")
    FOLDER_ZIP = ParamsVarEnv(key="NYSMIX_FOLDER_ZIP")
    FOLDER_CSV = ParamsVarEnv(key="NYSMIX_FOLDER_CSV")
    TYPE_DB = ParamsVarEnv(key="NYSMIX_DB")


from pandera import SchemaModel, Field, dataframe_check
from pandera.typing import Series, DataFrame
from datetime import datetime, date
from .concepts import Fuel, NAME
from typing import Optional


class Snapshot(SchemaModel):
    timestamp: Series[datetime] = Field(alias="Time Stamp", coerce=True)
    timezone: Series[str] = Field(isin=["EST", "EDT"], alias="Time Zone")
    fuel: Series[str] = Field(isin=[NAME[fuel] for fuel in Fuel], alias="Fuel Category")
    gen_mwh: Optional[Series[float]] = Field(coerce=True, alias="Gen MWh", ge=0)
    gen_mw: Optional[Series[float]] = Field(coerce=True, alias="Gen MW", ge=0)

    @dataframe_check
    def has_one_gen_mw_field(self, df: DataFrame) -> bool:
        return len(set(df.columns).intersection({self.gen_mw, self.gen_mwh})) == 1


class Summary(SchemaModel):
    date: Series[date] = Field(coerce=True)
    fuel: Series[str] = Field(isin=[NAME[fuel] for fuel in Fuel], alias="Fuel Category")
    gen_mw: Series[float] = Field(coerce=True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nysmix']

package_data = \
{'': ['*']}

install_requires = \
['joblib>=1.1.0,<2.0.0', 'pandera>=0.13.4,<0.14.0', 'pydantic>=1.10.1,<2.0.0']

extras_require = \
{'book': ['jupyter-book>=0.13.0,<0.14.0',
          'jupytext>=1.14.0,<2.0.0',
          'altair>=4.2.0,<5.0.0',
          'autodocsumm>=0.2.8,<0.3.0'],
 'google': ['joblibgcs>=0.2.0,<0.3.0'],
 'test': ['pytest>=7.1.2,<8.0.0', 'hypothesis>=6.61.0,<7.0.0']}

setup_kwargs = {
    'name': 'nysmix',
    'version': '0.6.0',
    'description': 'pulling nys fuel mix data from nysiso',
    'long_description': None,
    'author': 'Gautam Sisodia',
    'author_email': 'gautam.sisodia@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""
helpful abstractions
"""

from abc import ABC, abstractproperty
from dataclasses import dataclass

from joblib import Memory
from pandera.typing import DataFrame

from .schema import Summary


@dataclass(frozen=True)  # type: ignore
class TimeUnit(ABC):
    memory: Memory = Memory()

    @abstractproperty
    def last_modified(self) -> str:
        pass

    @abstractproperty
    def summary(self) -> DataFrame[Summary]:
        pass


class HasURL(ABC):
    @abstractproperty
    def url(self) -> str:
        pass

class HasZip(ABC):
    @abstractproperty
    def zip(self) -> bytes:
        pass
